import os
import json
import logging
from datetime import datetime
import tempfile
from collections import defaultdict
from pathlib import Path
import string
import socket

from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import NothingToReportException

from typing import List, Any, Dict, Optional, Callable

from db2_imports import connector
from db2_imports.models import QueryConfig, Result, QueryResult
from db2_imports.queries import queries
from db2_imports.log_proxy import LogProxy

# TODO:
# 3. Alerts?
# 4. Metrics from DB Insights
# 5. States for tablespace
# 6. Multi dimension views of metrics (aka charts)

# DONE:
# 12. Group name is host now
# 2. change absolute counters to relative
# 11. Help texts added to every metric chart.
# 10. Lock wait time normalized to 1000 transactions.
# 7. Name metrics properly ?wtf
# 9. Proper cache usage percentage calculation
# 13. Figure out better way to report availability of DB vs errors in ncalculation
# 14. Control metrics from configuration page?


logger = logging.getLogger(__name__)


class RemoteDB2Plugin(RemoteBasePlugin):
    def initialize(self, **kwargs):
        self.temp_folder = create_or_get_temp_folder()
        self.metrics_by_query_name: Dict[str, List[dict]] = self.metrics_by_query_name()

        # Write all queries to the config file
        member_scope = -1 if self.config["single_member"] else -2
        for query_config in queries:
            query_config.query = query_config.query.format(member_scope=member_scope)
        self.write_config(queries)

    def metrics_by_query_name(self) -> Dict[str, List[dict]]:
        # Here we group the metric configuration by query name
        # We need this to later resolve formulas and report type (absolute, relative, per_second)
        results = defaultdict(list)
        for metric in self.plugin_info.json_data.get("metrics", []):
            query_name = metric.get("source", {}).get("queryName", "")
            results[query_name].append(metric)
        return results

    def query(self, **kwargs):
        start_time = datetime.now()

        self.logger.setLevel(self.config.get("log_level", "INFO"))

        # Spawn the java process that will run the queries
        # The password is passed as an Environment Variable for security
        env = os.environ.copy()
        env["DATABASE_PASSWORD"] = self.config["database_password"]

        jvm_extra_args = []
        if self.config.get("jvm_extra_args"):
            jvm_extra_args = self.config.get("jvm_extra_args").splitlines()

        connector.query(self.activation.entity_id, self.temp_folder, env=env, java_path=self.config.get("jvm_path"), jvm_extra_args=jvm_extra_args)
        # Create the group and custom device for metric reporting
        group_name = f"{self.config['database_host']}:{self.config['database_port']}"
        group = self.topology_builder.create_group(group_name, group_name)
        self.custom_device = group.create_device(self.config["database_name"])
        self.report_endpoint()

        # Parse the results of the queries
        result = self.read_results_from_file()
        self.logger.info(f"Processing result {result}")

        total_reported_metrics = 1  # We always report availability

        if result.error:
            # This means there was an error connecting to the Database
            # Report the error and the availability
            error_message = f"Error running queries for {group_name}: {result.error_message}"
            self.logger.error(error_message)
            self.custom_device.report_error_event(error_message, "IBM DB2 Extension error")
            self.custom_device.absolute("AvailabilityPercent", 0)

        else:
            # If we get here we have succesfully queried the database, parse the results
            self.custom_device.absolute("AvailabilityPercent", 100)
            for query_result in result.query_results:
                self.logger.info(f"Processing query result {query_result}")

                # Grab all the metrics that depend on this query from plugin.json
                metrics_configs = self.metrics_by_query_name.get(query_result.name, [])
                total_reported_metrics += self.report_metrics(metrics_configs, query_result)

        self.parse_logs()
        time_taken = (datetime.now() - start_time).total_seconds()
        self.logger.info(f"Reported {total_reported_metrics} metrics to Dynatrace, query method time: {time_taken:.2f}s")

    def report_metrics(self, metric_configs: List[dict], query_result: QueryResult) -> int:
        total_reported_metrics = 0
        for metric_config in metric_configs:
            try:
                total_reported_metrics += self.process_metric(metric_config, query_result)
            except Exception as e:
                error_message = f"Could not process metric {metric_config} for query result {query_result}: {e}"
                self.logger.exception(error_message)
                self.custom_device.report_custom_info_event(error_message, "IBM DB2 Extension - Error processing metric")

        return total_reported_metrics

    def process_metric(self, metric_config: dict, query_result: QueryResult) -> int:
        total_reported_metrics = 0

        # First lets figure out what is the metric key
        # If we have dimensions, a formula and a metric_type
        # This list has already been filtered for only metrics that match the query name from query_result
        # All of this comes from plugin.json

        # Use the correct key depending on it being a normal or a state metric
        timeseries_key = "timeseries" if "timeseries" in metric_config else "statetimeseries"

        key = metric_config[timeseries_key]["key"]
        expected_dimensions = metric_config[timeseries_key].get("dimensions", [])
        formula = metric_config.get("source", {}).get("formula", "")  # This exists for every metric, checks just to be safe
        metric_type = metric_config.get("source", {}).get("type", "absolute")
        self.logger.debug(f"Processing metric {key} with formula: '{formula}' and type '{metric_type}'")

        # Trick to extract fields from a format string
        # From a string like '{rows_read} / ({rows_returned} + 1)'
        # This will generate ['rows_read', 'rows_returned']
        column_names = [v[1] for v in string.Formatter().parse(formula) if v[1] is not None]

        for row in query_result.rows:

            # We keep a key: value dictionary where the key is the column name
            # and the value is the column value, this will be used to format the formula
            # giving us the raw value, example a formula '{total_app_commits} + {total_app_rollbacks}'
            # generates a dict like '{'total_app_commits': '2531', 'total_app_rollbacks': '445'}'
            # which can be used to transform the formula to '2531 + 445'
            format_dict = {}
            for column_name in column_names:
                format_dict[column_name] = row.get_column_by_name(column_name).value
            resolved_formula = formula.format(**format_dict)

            # Danger zone, if plugin.json had the formula 'sys.exit()' we would be screwed
            # Todo - Check that the resulting formula only consists of numbers and valid math operators
            # Todo - Preferably without using numpy
            value = eval(resolved_formula)
            self.logger.debug(
                f"For metric {key} with formula: '{formula}' "
                f"extracted column names {column_names} and values {format_dict}, "
                f"the resulting formula is '{resolved_formula}' = '{value}'"
            )

            # Extract dimensions if they exist, also a dictionary
            # example: {'member': '0'}
            dimensions = {}
            for expected_dimension in expected_dimensions:
                dimensions[expected_dimension] = row.get_column_by_name(expected_dimension).value

            # The state must match the case defined in plugin.json
            # Ie if we get 'ACTIVE', send it as 'Active'
            # Don't try to be smart and change casing, just compare with
            # what is present in plugin.json and use that if we find a match
            for state in metric_config[timeseries_key].get("states", []):
                if str(value).upper() == state.upper():
                    self.logger.debug(f"The value from the database '{value}' will be sent as state '{state}'")
                    value = state

            # Exceptional cases that won't be handled by formulas
            if key == "LastBackupSecond":
                self.logger.debug(f"Calculating last backup from {value}")
                value = (datetime.now() - datetime.strptime(value, "%Y-%m-%d-%H:%M:%S")).total_seconds()

            # Only report metrics that the user asked for
            if (key in self.config and self.config[key]) or key not in self.config:
                total_reported_metrics += 1
                self.report_metric(key, value, dimensions, metric_type)

        return total_reported_metrics

    def report_metric(self, key: str, value: Any, dimensions: Optional[Dict[str, str]] = None, metric_type: str = "absolute"):
        # Get the method based on the name of the metric type
        # Ie, 'absolute' gives us the custom_device.absolute method
        method: Callable = self.custom_device.__getattribute__(metric_type)
        self.logger.debug(f"Calling {method} with key={key}, value={value}, dimensions={dimensions}")
        method(key, value, dimensions)

    def build_connection_string(self) -> str:
        # Check if the user supplied a connection string
        connection_string = self.config.get("connection_string")
        if not connection_string:
            connection_string = "jdbc:db2://{host}:{port}/{database}{use_ssl}".format(
                host=self.config["database_host"],
                port=self.config["database_port"],
                database=self.config["database_name"],
                use_ssl=":sslConnection=true;" if self.config["use_ssl"] else "",
            )
        self.logger.info(f"Using connection string {connection_string}")
        return connection_string

    def write_config(self, queries_to_run: List[QueryConfig]):
        """
        Writes the database configuration to a json file named after the endpoint ID
        """
        queries_json = [query.__dict__ for query in queries_to_run]
        config = {
            "connectionString": self.build_connection_string(),
            "username": self.config["database_user"],
            "queries": queries_json,
            "driverClassName": "com.ibm.db2.jcc.DB2Driver",
        }

        path = Path(self.temp_folder, "config")
        if not path.exists():
            os.mkdir(path)
        with open(Path(path, f"{self.activation.entity_id}.json"), "w") as f:
            json.dump(config, f, indent=2)

    def read_results_from_file(self) -> Result:
        """
        Reads the <endpoint_id>.json file inside the results folder
        and parses the data into a models.Result object
        """
        results_file = Path(self.temp_folder, "results", f"{self.activation.entity_id}.json")
        if results_file.exists():
            with open(results_file, "r") as f:
                try:
                    return Result.from_json(json.load(f))
                except Exception as e:
                    self.logger.error(f"Could not parse json from {results_file}: {e}")
                    raise NothingToReportException(f"Error: {e}")
        return Result()

    def parse_logs(self):
        """
        Simply logs all lines from the java process logs to the extension logs
        """
        log = Path(self.temp_folder, "log", f"{self.activation.entity_id}.log")
        cache = Path(self.temp_folder, f"{self.activation.entity_id}.json")

        for line in LogProxy(log, cache).get_lines_from_file():
            self.logger.info(line)

    def report_endpoint(self):
        try:
            ip = socket.gethostbyname(self.config["database_host"])
            self.custom_device.add_endpoint(ip, self.config["database_port"], dnsNames=[self.config["database_host"]])
        except Exception as e:
            self.logger.warning(f"Could not determine the host IP address: {e} to report an endpoint")


def create_or_get_temp_folder() -> Path:
    temp_folder = Path(tempfile.gettempdir(), "dynatrace-db2-extension")
    if not temp_folder.exists():
        os.mkdir(temp_folder)
    return temp_folder
